import hashlib
import pickle

import doit.action

def _command_digest(action):
    hash = hashlib.sha1()
    for item in action.action:
        hash.update(str(item).encode())
    return hash.hexdigest()

def _python_digest(action):
    hash = hashlib.sha1()
    hash.update(action.py_callable.__code__.co_code)
    hash.update(pickle.dumps(action.args))
    hash.update(pickle.dumps(action.kwargs))
    return hash.hexdigest()

_digest = {
    doit.action.CmdAction: _command_digest,
    doit.action.PythonAction: _python_digest
}

def _check(task, values):
    digests = [_digest[type(action)](action) for action in task.actions]
    task.value_savers.append(lambda: {"actions_hash": digests})
    return values.get("actions_hash", None) == digests

def uptodate(self):
    return [_check]

class classproperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, _, class_):
        return self.fget(class_)



import os

from .task_factory import TaskFactory

class Registration(TaskFactory):
    """ Register two images using ANTs. fixed and moving describe the respective
        images, and may include an optional volume to use in case of 4D images:
        passing (foo.nii.gz,0) will use the first 3D volume of the 4D foo.nii.gz
        for the registration. transform must be one of "rigid", "affine" or 
        "syn". initial_transforms must be either None (its default value) or a
        list of transforms.
        
        This task stores the transforms in a specific member, in the order they
        should be passed to ApplyTransforms.
    """
    def __init__(
            self, fixed, moving, transform, prefix, 
            save_warped=True, quick=False, precision="double", 
            initial_transforms=None):
        TaskFactory.__init__(self, str(prefix))
        self.quick = quick
        
        # Prepare the volume extraction if necessary
        self.file_dep = []
        volumes = []
        extractions = []
        removals = []
        for name, data in [["fixed", fixed], ["moving", moving]]:
            path, index = _get_path_and_index(data)
            self.file_dep.append(path)
            
            if None not in (path, index):
                # NOTE Named is based on target, it should be unique since 
                # target paths are.
                temp = os.path.join(
                    os.path.dirname(prefix), 
                    "__{}_{}_{}.nii.gz".format(
                        os.path.basename(prefix), name, index))
                
                volumes.append(temp)
                extractions.append(
                    ["ImageMath", "4", temp, "ExtractSlice", path, str(index)])
                removals.append(["rm", temp])
            else:
                volumes.append(path)
        
        fixed_volume, moving_volume = volumes
        
        # Prepare the registration command
        registration = [
            "antsRegistration",
            "--dimensionality", "3", 
            "--float", "0" if precision == "double" else "1",
            "--interpolation", "Linear", 
            "--winsorize-image-intensities", "[0.005,0.995]",
            "--use-histogram-matching", "0",
        ]
        
        # Update the outputs of the command
        output_images = []
        if save_warped:
            registration += [
                "--output", 
                "[{},{}Warped.nii.gz,{}InverseWarped.nii.gz]".format(
                    prefix, prefix, prefix)]
            output_images += [
                "{}Warped.nii.gz".format(prefix), 
                "{}InverseWarped.nii.gz".format(prefix)]
        else:
            registration += ["--output", prefix]
        
        # Update the command with the transforms. WARNING: antsApplyTransforms
        # uses a transform _stack_.
        self.transforms = []
        registration += self.initial_stage(
            fixed_volume, moving_volume, initial_transforms)
        if transform.lower() in ["rigid", "affine", "syn"]:
            registration += self.rigid_stage(fixed_volume, moving_volume)
            self.transforms.insert(0, "{}{}".format(prefix, "0GenericAffine.mat"))
        if transform.lower() in ["affine", "syn"]:
            registration += self.affine_stage(fixed_volume, moving_volume)
        if transform.lower() == "syn":
            registration += self.syn_stage(fixed_volume, moving_volume)
            self.transforms.insert(0, "{}{}".format(prefix, "1Warp.nii.gz"))
        
        # self.file_dep is already OK
        self.targets = (
            self.transforms 
            + (
                ["{}{}".format(prefix, "1InverseWarp.nii.gz")] 
                if transform.lower() == "syn" else [])
            + output_images)
        self.actions = extractions + [registration] + removals
        
    @property
    def inverse_transforms(self):
        """ The list of transforms from fixed to moving, in the order they 
            should be passed to ApplyTransforms.
        """
        
        result = []
        for transform in self.transforms[::-1]:
            if transform.endswith("0GenericAffine.mat"):
                result.append([transform, 1])
            else:
                result.append(transform.replace("Warp.nii", "InverseWarp.nii"))
        return result
    
    def initial_stage(self, fixed, moving, initial_transforms):
        if not initial_transforms:
            initial_transforms = ["[{},{},1]".format(fixed, moving)]
        transforms = []
        for item in initial_transforms:
            transforms.extend(["--initial-moving-transform", item])
        return transforms
    
    def rigid_stage(self, fixed, moving):
        return [
            "--transform", "Rigid[0.1]",
            "--metric", "MI[{},{},1,32,Regular,0.25]".format(fixed, moving),
            "--convergence", "[1000x500x250x{},1e-6,10]".format(
                0 if self.quick else 100),
            "--shrink-factors", "8x4x2x1",
            "--smoothing-sigmas", "3x2x1x0vox",
        ]
    
    def affine_stage(self, fixed, moving):
        return [
            "--transform", "Affine[0.1]",
            "--metric", "MI[{},{},1,32,Regular,0.25]".format(fixed, moving),
            "--convergence", "[1000x500x250x{},1e-6,10]".format(
                0 if self.quick else 100),
            "--shrink-factors", "8x4x2x1",
            "--smoothing-sigmas", "3x2x1x0vox",
        ]
    
    def syn_stage(self, fixed, moving):
        return [
            "--transform", "SyN[0.1,3,0]",
            "--metric", "CC[{},{},1,4]".format(fixed, moving),
            "--convergence", "[100x70x50x{},1e-6,10]".format(
                0 if self.quick else 20),
            "--shrink-factors", "8x4x2x1",
            "--smoothing-sigmas", "3x2x1x0vox",
        ]

class ApplyTransforms(TaskFactory):
    """ Apply transforms and resample an image. The reference image may include
        an optional volume to use in case of 4D images: passing (foo.nii.gz,0) 
        will use the first 3D volume of the 4D foo.nii.gz.
    """
    def __init__(
            self, input, reference, transforms, output, 
            interpolation="BSpline", input_image_type="scalar"):
        TaskFactory.__init__(self, str(output))
        
        extraction = []
        removal = []
        
        reference_path, index = _get_path_and_index(reference)
        
        if None not in (reference_path, index):
            # NOTE Named is based on target, it should be unique since 
            # target paths are.
            reference_volume = os.path.join(
                os.path.dirname(output), 
                "__{}_{}.nii.gz".format(os.path.basename(output), index))
        
            extraction = [
                "ImageMath", "4", reference_volume,
                "ExtractSlice", reference_path, str(index)]
            removal = ["rm", reference_volume]
        else:
            reference_volume = reference_path
        
        self.file_dep = [input, reference_path]
        for transform in transforms:
            if isinstance(transform, list):
                self.file_dep.append(transform[0])
            else:
                self.file_dep.append(transform)
        
        self.targets = [output]
        apply_transforms = [
            "antsApplyTransforms",
            "-i", input, "-r", reference_volume, "-o", output, 
            "-n", interpolation, "-e", input_image_type]
        for transform in transforms:
            apply_transforms.append("-t")
            if isinstance(transform, (list, tuple)):
                apply_transforms.append("[{}]".format("{},{}".format(*transform)))
            else:
                apply_transforms.append(transform)
        self.actions = extraction+[apply_transforms]+removal

def _get_path_and_index(data):
    if isinstance(data, (list, tuple)):
        path, index = data
    else:
        path, index = data, None
    return path, index



import itertools
import textwrap

import numpy

from .spm_object import SPMObject

class Contrast(object):
    """ T-constrast of F-contrast.
    """
    
    def __init__(self, name, weights, replication="none"):
        self.name = name
        self.weights = numpy.asarray(weights)
        if len(self.weights.shape) > 1:
            raise NotImplementedError("F-contrast not implemented")
        
        if replication not in ["none", "repl", "replsc", "sess", "both", "bothsc"]:
            raise NotImplementedError(
                "Unknown replication mode: {}".format(replication))
        self.replication = replication

class ContrastManager(SPMObject):
    """ Specify T and F contrasts
    """
    
    @staticmethod
    def pairwise(design, levels):
        """ Return all pairwise contrasts of factor with specified levels.
        """
        
        permutations = itertools.permutations(range(len(levels)), 2)
        contrasts = []
        for x,y in permutations:
            name = "{} < {}".format(levels[x], levels[y]) 
            
            weights = [0]*len(levels)
            weights[x] = -1
            weights[y] = +1
            
            contrasts.append(Contrast(name, weights))
        
        return ContrastManager(design, contrasts)
    
    def __init__(self, design, contrasts=None):
        SPMObject.__init__(self, "spm.stats.con")
        
        self.design = design
        self.contrasts = contrasts or []
        
        self.template = textwrap.dedent("""\
            {{ id(index, name) }}.spmmat = {'{{ design.spmmat }}'};
            {% for contrast in contrasts -%}
            {{ id(index, name) }}.consess{{ "{"+(loop.index|string)+"}" }}.tcon.name = '{{ contrast.name }}';
            {{ id(index, name) }}.consess{{ "{"+(loop.index|string)+"}" }}.tcon.weights = [{{ contrast.weights|join(" ") }}];
            {{ id(index, name) }}.consess{{ "{"+(loop.index|string)+"}" }}.tcon.sessrep = '{{ contrast.replication }}';
            {% endfor -%}
            {{ id(index, name) }}.delete = 1;""")
    
    def _get_file_dep(self):
        return [self.design.spmmat]
    
    def _get_targets(self):
        targets = [self.design.spmmat]
        
        directory = self.design.spmmat.parent
        for index, contrast in enumerate(self.contrasts):
            type_ = "F" if numpy.squeeze(contrast.weights).ndim > 1 else "T"
            targets.extend([
                directory/"con_{:04d}.nii".format(1+index),
                directory/"spm{}_{:04d}.nii".format(type_, 1+index)])
        return targets



import inspect

import doit

from .task_factory import TaskFactory

class TaskGraph(object):
    def __init__(self, tasks):
        # List of children for each Doit task
        self.children = {}
        # All Doit tasks (Spire and non-Spire)
        self.doit_tasks = {task.name: task for task in tasks}
        # Spire-only tasks
        self.spire_tasks = {
            task.basename: task for task in TaskFactory._task_registry}
        
        # NOTE: this is O(n^2).
        for task in tasks:
            children = []
            for other_task in tasks:
                if any(x in task.targets for x in other_task.file_dep):
                    children.append(other_task.name)
            
            self.children[task.name] = children



import jinja2

class SPMObject(object):
    """ Abstract Base Class for all SPM objects.
        
        Even though SPM objects are not Spire tasks (as some of them will modify
        in-place their file_dep, which is not compatible with doit's task 
        semantics), they nonetheless include task-related properties: file_dep
        and targets. Subclasses will have to override the _get_file_dep and
        _get_targets functions to return the correct values.
    """
    
    def __init__(self, name):
        self.name = name
        self.environment = jinja2.Environment()
        self.environment.globals.update(id=__class__._get_id)
    
    def get_script(self, index):
        template = self.environment.from_string(self.template)
        return template.render(index=index, **vars(self))
    
    @property
    def file_dep(self):
        return self._get_file_dep()
    
    @property
    def targets(self):
        return self._get_targets()
    
    @staticmethod
    def _get_id(index, name):
        return "matlabbatch{"+str(index)+"}."+name
    
    def _get_file_dep(self):
        return []
    
    def _get_targets(self):
        return []
    
    def __getstate__(self):
        state = self.__dict__.copy()
        del state["environment"]
        return state
    
    def __setstate__(self, state):
        self.__dict__.update(state)
        self.environment = jinja2.Environment()
        self.environment.globals.update(id=__class__._get_id)



from . import contrast_manager
from . import factorial_design
from .image_calculator import ImageCalculator
from .model_estimation import ModelEstimation
from .spm_object import SPMObject
from .stats import Stats
from .utils import *



from . import misc

class TaskFactory(object):
    """ Base class for task factory objects. Derived classes must follow this
        example:
        
        >>> class MyTask(TaskFactory):
        ...     def __init__(self, input, output):
        ...         # Initialize base class
        ...         TaskFactory.__init__(self, "my basename")
        ...         
        ...         # Define mandatory members
        ...         self.file_dep = [input]
        ...         self.targets = [output]
        ...         self.actions = ["touch {}".format(output)]
        
        A task can then be created as such:
        >>> task = MyTask("foo", "bar")
        
        If the task is not re-used, storing it in an object is not mandatory.
    """
    
    _task_registry = []
    
    def __init__(self, basename):
        self.basename = basename
        self.clean = True
        TaskFactory._task_registry.append(self)
    
    def as_task_dict(self):
        if getattr(self, "skipped", False):
            return None
        else:
            fields = ["basename", "file_dep", "targets", "actions", "clean", "uptodate"]
            return {x: getattr(self, x) for x in fields}
    
    @property
    def target(self):
        if len(self.targets) == 1:
            return self.targets[0]
        else:
            raise Exception(
                "Ambiguous call to \"target\" ({} target{})".format(
                    len(self.targets), "" if len(self.targets)==0 else "s"))
    
    uptodate = property(misc.uptodate)



import glob
import os

from setuptools import setup, find_packages

here = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(here, "README.md")) as f:
    long_description = f.read()

setup(
    name="spire-pipeline",
    version="1.2.0",
    
    description="Run software pipelines using doit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    
    url="https://github.com/lamyj/spire",
    
    author="Julien Lamy",
    author_email="lamy@unistra.fr",
    
    license="MIT",
    
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        
        "Environment :: Console",
        
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        
        "Topic :: Software Development :: Build Tools",
        "Topic :: Scientific/Engineering",
        
        "License :: OSI Approved :: MIT License",
        
        "Programming Language :: Python :: 3",
    ],
    
    keywords="pipeline, workflow, task execution",

    packages=find_packages(exclude=["tests"]),
    install_requires=["doit", "jinja2", "numpy"],
)



import argparse
import os
import subprocess
import sys
import tempfile

def graph(tasks_path, dot_path, tasks_only, doit_arguments):
    temp_fd, temp_path = tempfile.mkstemp(suffix=".py")
    try:
        with open(tasks_path) as fd:
            os.write(temp_fd, fd.read().encode())
        os.write(
            temp_fd, 
            "\n"
            "import spire\n"
            "graph = spire.graph({})\n"
            "with open(\"{}\", \"w\") as fd:\n"
            "    fd.write(graph)\n".format(tasks_only, dot_path).encode())
        os.close(temp_fd)
        subprocess.check_call(["doit", "list", "-f", temp_path]+doit_arguments)
    finally:
        os.remove(temp_path)

if __name__ == "__main__":
    action = sys.argv[1]
    if action == "graph":
        parser = argparse.ArgumentParser(
            description="Generate a Graphviz representation of the task graph")
        parser.add_argument(
            "tasks_path", metavar="tasks.py", help="Path to the task graph")
        parser.add_argument(
            "dot_path", metavar="tasks.dot", help="Path to the task graph")
        parser.add_argument(
            "--tasks-only", action="store_true", help="Create nodes for tasks only")
        arguments, doit_arguments = parser.parse_known_args(sys.argv[2:])
        
        arguments.dot_path = os.path.abspath(arguments.dot_path)
        
        sys.exit(graph(doit_arguments=doit_arguments, **vars(arguments)))
        



from . import misc

class TaskMeta(type):
    def __new__(class_, name, bases, dict_):
        if bases[0] is not TaskBase:
            if "basename" not in dict_:
                dict_["basename"] = name
            
            for singular, plural in [["action", "actions"], ["target", "targets"]]:
                if plural not in dict_:
                    if singular not in dict_:
                        raise Exception(
                            "Neither \"{}\" nor \"{}\" exist".format(
                                singular, plural))
                    else:
                        dict_[plural] = [dict_[singular]]
                elif singular not in dict_:
                    if len(dict_[plural]) == 1:
                        dict_[singular] = dict_[plural][0]
            
            if "file_dep" not in dict_:
                raise Exception("No \"file_dep\" specified")
            elif not isinstance(dict_["file_dep"], (list, tuple)):
                dict_["file_dep"] = [dict_["file_dep"]]
            
        return type.__new__(class_, name, bases, dict_)

# Base class with the correct meta-class. Required to keep compatibility between
# Python 2 and Python 3; borrowed from six.
# https://bitbucket.org/gutworth/six/src/92e1c74/six.py?at=default&fileviewer=file-view-default#six.py-815:824
TaskBase = type.__new__(TaskMeta, "TaskBase", (), {})

class Task(TaskBase): 
    clean = True

    @misc.classproperty
    def uptodate(cls):
        return misc.uptodate(cls)

    @classmethod
    def create_doit_tasks(class_):
        if class_ is Task:
            return

        if getattr(class_, "skipped", False):
            return None
        else:
            fields = ["basename", "file_dep", "targets", "actions", "clean", "uptodate"]
            return {x: getattr(class_, x) for x in fields}



import itertools

from .. import TaskFactory
from . import model_estimation, utils

class Stats(TaskFactory):
    """ Perform statistical analyses with SPM using given design and contrasts.
    """
    
    def __init__(self, name, design, contrasts):
        TaskFactory.__init__(self, name)
        
        estimation = model_estimation.ModelEstimation(design)
        
        steps = [design, estimation, contrasts]
        
        self.file_dep = design.file_dep
        self.targets = list(set(itertools.chain(*[x.targets for x in steps])))
        self.actions = [
            # Make sure SPM.mat is not there when we start
            ["rm", "-f", design.spmmat],
            (utils.run, (steps,))
        ]



import inspect
import itertools
import logging

import doit

from .task import Task
from .task_factory import TaskFactory
from .task_graph import TaskGraph

################################################################################
# Replace doit's task loader to include objects created by TaskFactory         #
################################################################################

doit_loader_load_tasks = None
def spire_load_tasks(*args, **kwargs):
    tasks = doit_loader_load_tasks(*args, **kwargs)
    
    for object_ in TaskFactory._task_registry:
        if getattr(object_, "skipped", False):
            continue
        
        dict_ = object_.as_task_dict()
        if dict_ is None:
            continue
        
        dict_["name"] = dict_.pop("basename")
        # Replace None with "" to avoid an error on task creation
        dict_["file_dep"] = [x or "" for x in dict_["file_dep"]]
            
        tasks.append(doit.task.dict_to_task(dict_))
            
    return tasks
    
if doit.loader.load_tasks != spire_load_tasks:
    doit_loader_load_tasks = doit.loader.load_tasks
    doit.loader.load_tasks = spire_load_tasks

################################################################################
# Prune the task graph from tasks having None in their file_dep                #
################################################################################

def prune():
    caller = inspect.getouterframes(inspect.currentframe())[1][0]
    tasks = doit.loader.load_tasks(caller.f_globals)
    
    graph = TaskGraph(tasks)
    
    to_skip = []
    for name, doit_task in graph.doit_tasks.items():
        if "" in doit_task.file_dep:
            to_skip.append(name)
    
    while len(to_skip) > 0:
        root = to_skip.pop()
        spire_task = graph.spire_tasks.get(root)
        if spire_task is None:
            logging.info("Not a spire task: {}".format(root))
            continue
        if not getattr(spire_task, "skipped", False):
            logging.warning("Skipping {}".format(root))
            spire_task.skipped = True
            to_skip.extend(graph.children.get(root, []))

################################################################################
# Representation of the task graph in the Graphviz format                      #
################################################################################

def graph(name_mapper=None):
    caller = inspect.getouterframes(inspect.currentframe())[1][0]
    tasks = doit.loader.load_tasks(caller.f_globals)
    
    graph = TaskGraph(tasks)
    
    if name_mapper is None:
        name_mapper = lambda x: x
    
    def quote(name):
        return "\"{}\"".format(name_mapper(name).replace("\"", "\\\""))
    
    lines = ["digraph {"]
    
    file_deps = set()
    targets = set()
    nodes = set()
    for task in graph.doit_tasks.values():
        file_deps.update(task.file_dep)
        targets.update(task.targets)
    for name, doit_task in graph.doit_tasks.items():
        if name not in nodes:
            lines.append("    {}[shape=box];".format(quote(name)))
        for entry in graph.doit_tasks[name].file_dep:
            if entry not in targets:
                if entry not in nodes:
                    lines.append(
                        "    {}[shape=box,color=blue];".format(quote(entry)))
                    nodes.add(entry)
                lines.append("    {} -> {};".format(quote(entry), quote(name)))
        for entry in graph.doit_tasks[name].targets:
            if entry not in file_deps and entry != name:
                if entry not in nodes:
                    lines.append(
                        "    {}[shape=box,color=blue];".format(quote(entry)))
                    nodes.add(entry)
                lines.append("    {} -> {};".format(quote(name), quote(entry)))
    
    for name, doit_task in graph.doit_tasks.items():
        for child in graph.children[name]:
            lines.append("    {} -> {};".format(quote(name), quote(child)))
    
    lines.extend(["}", ""])
    
    return "\n".join(lines)
